#!/usr/bin/env python 
# -*- coding: UTF-8 -*-
# ----------------------
# Name: ttvdb.py
# Python Script
# Author: 	R.D. Vaughan
# Purpose: 	This python script is intended to perform TV series data lookups based on information found on
# 			the http://thetvdb.com/ website. It follows the MythTV standards set for the movie data
#           lookups. e.g. the perl script "tmdb.pl" used to access themoviedb.com
#			This script uses the python module tvdb_api.py (v0.5 or higher) found at 
#			http://pypi.python.org/pypi?%3Aaction=search&term=tvnamer&submit=search thanks 
#			to the authors of this excellant module.
#			The tvdb_api.py module uses the full access XML api published by thetvdb.com see:
#			http://thetvdb.com/wiki/index.php?title=Programmers_API
#			Users of this script are encouraged to populate thetvdb.com with TV show information, posters,
#			fan art and banners. The richer the source the more valuable the script.
#			This python script was modified based on the "tvnamer.py" created by "dbr/Ben" who is also
#           the author of the "tvdb_api.py" module. "tvnamer.py" is used to rename avi files with 
#           series/episode information found at thetvdb.com
# Command example:
# See help (-u and -h) options
#
# Design:
#	1) Verify the command line options (display help or version and exit)
#	2) Verify that thetvdb.com has the series or series_season_ep being requested exit if does not exit
#	3) Find the requested information and send to stdout if any found
#
#
# License:Creative Commons GNU GPL v2
# (http://creativecommons.org/licenses/GPL/2.0/)
#-------------------------------------
__title__ ="thetvdb.com Query Engine";
__author__="R.D.Vaughan"
__version__="v0.8.8"	 	# Version .1 Initial development 
							# Version .2 Add an option to get season and episode numbers from ep name 
							# Version .3 Cleaned up the documentation and added a usage display option
							# Version .4 Added override formating of the number option (-N) 
							# Version .5 Added a check that tvdb_api.py dependancies are installed  
							# Version .6 Added -M Series List functionality  
							# Version .7 Added -o Series name override file functionality  
							# Version .8 Removed a dependancy, fixed caching for multiple users and
							#			 used better method of supporting the -M option with tvdb_api
							# Version .8.1 Cleaned up documentation errors.
							# Version .8.2 Added the series name to output of meta data -D
							# Version .8.3 Patched tv_api so even episode image is fully qualified URL
							# Version .8.4 Fixed bug in -N when multiple episodes are returned from
							#              a search on episode name.
							# Version .8.5 Made option -N more flexible in finding a matching episode
							# Version .8.6 Add option -t for top rated graphics (-P,B,F) for a series
							#              Add option -l to filter graphics on the local language 
							#              Add season level granularity to Poster and Banner URLs 
							#              Changed the override file location to be specified on the 
                            #              command line along with the option -o.
							#              Increased the amount of massaging of episode data to improve
							#			   compatiblilty with data bases. 
							#			   Changed the default season episode number format to SxxExx 
							#			   Add an option (-n) to return the season numbers for a series
 							#			   Added passing either a thetvdb.com SID (series identifcation
							#			   number) or series name for all functions except -M list. 							#			   Option -M always assumes a series name is used.
							#			   Now ALL available episode meta data is returned. This 
                            #              includes meta data that MythTv DB does not currently store.
                            #              Meta data 'Year' now derived from year in release date.
							# Version .8.7 Fixed a bug with unicode meta data handling
							# Version .8.8 Replaced the old configuration file with a true conf file
							#              Add option -m to better conform to mythvideo standards


usage_txt='''
This script fetches TV series information from theTVdb.com web site. The script conforms to MythTV's 
query engines for Movie Data standards.
NOTE: In the case of multiple queries the display order will always be:
    Numbers         - Used as a single option not valid as a multi-option parameter 
    List            - Used as a single option not valid as a multi-option parameter 
    Data            - The initial line of an episode's data starts with 'Series:'
    Poster URL(s)   - The initial line is 'poster:' then the URL(s) follow on the next line(s)
    Fan art URL(s)  - The initial line is 'fanart:' then the URL(s) follow on the next line(s)
    Banner URL(s)   - The initial line is 'banner:' then the URL(s) follow on the next line(s)
NOTE: Queries with no results have no output but have a return code of "True".
NOTE: All errors starts with a '! ' and a return code of "False". 
	
Options include:
  -h, --help        : Display help
  -u, --usage       : Examples on using this script
  -v, --version     : Displays the script title, version number and script author
  -d, --debug       : Verbose, debug information
  -r, --raw         : Display raw XML information returned from various queries
  -i, --interactive : Allows command line interaction when selecting from a number of possible series
                      Default selects the first series matching the series name so be specific.
  -c, --configure   : User configuration file, set series name overrides, episode name massaging and
                      regex strings for parsing video file names for series, season and episode numbers.
                      See the example file "override_example.txt" for details
  -l, --language    : Select only the graphics that match the users local language (e.g. 'en' English) 
  -t, --toprated    : Select only the top rated graphics for a TV Series, 
                      (Poster and/or Fanart and/or Banner)
  -n, --num_series  : Return the season numbers for a series 
  -m, --mythvideo   : Conform to mythvideo standards when processing -M, -P,
                      -F and -D options.
  -M, --list        : Returns a list of matching series
  -D, --data        : Returns the series episode information
  -B, --backdrop    : Returns the URL(s) to the series banner graphics
  -F, --fanart      : Returns the URL(s) to the series fan art graphics
  -P, --poster      : Returns the URL(s) to the series poster graphics
  -N, --numbers     : Returns the season number and episode number from an episode name 
 NOTE: for options (MPFBDN) an empty string is returned to stdout if no information was found

2nd/3rd/4th args   : (not required for -h, -v or -u) 
    One of the following combinations:
       Series name or SID (Series Identification number from thetvdb.com)
       Series name/SID Episode name
       Series name/SID Season number
       Series name/SID Season number Episode number
       A video episode file name which contains series name, season and episode numbers 
       (no file extention) e.g. "Stargate SG-1 - S08E03 - Lockdown"

Special notes:
    -B (banner graphics requests) can be either series specific or series/season specific 
    -P (poster graphics requests) can be either series specific or series/season specific 
    -F (fan art graphics requests) are only series specific 
    -D (episode data requests) Episode data for:
       "all seasons" - Only a Series name/SID is provided
       "a specific season" - A Series name/SID and season number are provided
       "a specific episode" - A Series name/SID and season number and episode number are provided
       "a specific episode" - A video episode file name which contains series name, season and episode 
                              numbers are provided
    -N (season and episode numbers) A Series name/SID and Episode name are provided
    NOTE: There are no identifiers for URLs output if ONLY one type has been selected (e.g -B versus -BF)
 
Command examples:
(Return the banner graphics for a series)
> ttvdb -B "Sanctuary"
http://www.thetvdb.com/banners/seasons/80159-1.jpg
http://www.thetvdb.com/banners/seasons/80159-1-2.jpg
http://www.thetvdb.com/banners/seasons/80159-1-4.jpg
http://www.thetvdb.com/banners/seasons/80159-1-3.jpg

(Return the banner graphics for a Series specific to a season)
> ttvdb -B "SG-1" 1
http://www.thetvdb.com/banners/seasonswide/72449-1.jpg

(Return the banner graphics for a SID (series ID) specific to a season)
(SID "72449" is specific for the series "SG-1")
> ttvdb -B 72449 1
http://www.thetvdb.com/banners/seasonswide/72449-1.jpg

(Return the banner graphics for a file name)
> ttvdb -B "Stargate SG-1 - S08E03 - Lockdown"
http://images.thetvdb.com.nyud.net:8080/banners/seasonswide/72449-8.jpg

(Return the posters, banners and fan art for a series)
> ttvdb -PFB "Sanctuary"
poster:
http://www.thetvdb.com/banners/posters/80159-1.jpg
fanart:
http://www.thetvdb.com/banners/fanart/original/80159-5.jpg
http://www.thetvdb.com/banners/fanart/original/80159-1.jpg
banner:
http://www.thetvdb.com/banners/seasons/80159-1.jpg

(Return thetvdb.com's top rated poster, banner and fan art for a TV Series)
(NOTE: If there is no graphic for a type or any graphics at all then those types are not returned)
> ttvdb -tPFB "Stargate SG-1"
Poster:http://www.thetvdb.com/banners/posters/72449-1.jpg
Fanart:http://www.thetvdb.com/banners/fanart/original/72449-1.jpg
Banner:http://www.thetvdb.com/banners/graphical/185-g3.jpg
> ttvdb -tB "Night Gallery"
http://www.thetvdb.com/banners/blank/70382.jpg

(Return graphics only matching the local language for a TV series)
(In this case banner 73739-g9.jpg is not included because it does not match the local language 'en')
> ttvdb -lB "Lost"
http://www.thetvdb.com/banners/graphical/24313-g.jpg
http://www.thetvdb.com/banners/graphical/73739-g6.jpg
...
http://www.thetvdb.com/banners/graphical/73739-g2.jpg
http://www.thetvdb.com/banners/blank/73739.jpg

(Return a season and episode numbers) Note: Any other options are ignored e.g. (-BFP ... etc)) 
> ttvdb -N "Eleventh Hour (US)" "H2O"
S01E10

(Return a season and episode numbers with an override of the default output format)
(WARNING: There is no editing of an override output format, two '%d' must be included)
> ttvdb -N "Eleventh Hour (US)" "H2O" "Season%dx%02d"
Season1x10

(Return a season and episode numbers using the override file to identify the series as the US version)
> ttvdb -N -configure="/home/user/.tvdb/tvdb.conf" "Eleventh Hour" "H2O"
S01E10

(Return the season numbers for a series)
> ttvdb -n -configure="/home/user/.tvdb/tvdb.conf" -n "SG-1"
0,1,2,3,4,5,6,7,8,9,10

(Return the meta data for a specific series/season/episode)
> ttvdb -D "Sanctuary" 0 1
Series:Sanctuary
Season:0
Episode:1
Title:Webisode 1
Year:2007
ReleaseDate:2007-05-14
Director:Martin Wood
Plot:While tracking a young boy with dangerous powers, Dr. Helen Magnus encounters Will Zimmerman, a psychiatric resident who may have what it takes to become her new protégé.
UserRating:8.0
Writers:Damian Kindler, Martin Wood
Episode Image:http://www.thetvdb.com/banners/episodes/80159-328953.jpg
Language:en
Combined_Episodenumber:1
Id:328953
Seasonid:27498
Lastupdated:1223234120
Combined_Season:0
Gueststars:Dan Payne, Sarah Sands, Lorena Gale, Sheri Noel, Diana Pavlovska, Malik McCall, Cainan Wiebe, Kavan Smith, Michael Adamthwaite
Seriesid:80159
Productioncode:101

(Return a list of "thetv.com series id and series name" that contain specific search word(s) )
(!! Be careful with this option as poorly defined search words can result in large lists being returned !!)
> ttvdb.py -M "night a"
74382:Love on a Saturday Night
71476:Star For A Night

'''
# Episode keys that can be used in a episode data/information search. 
# Not all keys are currently being used. 
'''
'episodenumber'
'rating'
'overview'
'dvd_episodenumber'
'dvd_discid'
'combined_episodenumber'
'epimgflag'
'id'
'seasonid'
'seasonnumber'
'writer'
'lastupdated'
'filename'
'absolute_number'
'combined_season'
'imdb_id'
'director'
'dvd_chapter'
'dvd_season'
'gueststars'
'seriesid'
'language'
'productioncode'
'firstaired'
'episodename'
'''


# System modules
import sys, os, re, locale, ConfigParser
from optparse import OptionParser

# Verify that tvdb_api.py, tvdb_ui.py and tvdb_exceptions.py are available
try:
	import tvdb_ui
	# thetvdb.com specific modules
	from tvdb_api import (tvdb_error, tvdb_shownotfound, tvdb_seasonnotfound, tvdb_episodenotfound, tvdb_episodenotfound, tvdb_attributenotfound, tvdb_userabort)
#	from tvdb_api import Tvdb
	import tvdb_api
except:
	print '''
The modules tvdb_api.py (v0.6.0 or greater), tvdb_ui.py, tvdb_exceptions.py and cache.py must be
in the same directory as ttvdb.py. They should have been included with the distribution of ttvdb.py.
'''
	sys.exit(False)

# Global variables
http_find="http://www.thetvdb.com"
http_replace="http://images.thetvdb.com.nyud.net:8080"

logfile="/home/doug/joblog/tvdb.log"

valid_filename_chars=u"""0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ!@£$%^&*()_+=-[]{}"'.,<>`~?: """
valid_filename_chars_regex = re.escape(valid_filename_chars)
name_parse=[
			# foo_[s01]_[e01]
			re.compile('''^([%s]+?)[ \._\-]\[[Ss]([0-9]+?)\]_\[[Ee]([0-9]+?)\]?[^\\/]*$'''% (valid_filename_chars_regex)),
			# foo.1x09*
			re.compile('''^([%s]+?)[ \._\-]\[?([0-9]+)x([0-9]+)[^\\/]*$''' % (valid_filename_chars_regex)),
			# foo.s01.e01, foo.s01_e01
			re.compile('''^([%s]+?)[ \._\-][Ss]([0-9]+)[\.\- ]?[Ee]([0-9]+)[^\\/]*$''' % (valid_filename_chars_regex)),
			# foo.103*
			re.compile('''^([%s]+)[ \._\-]([0-9]{1})([0-9]{2})[\._ -][^\\/]*$''' % (valid_filename_chars_regex)),
			# foo.0103*
			re.compile('''^([%s]+)[ \._\-]([0-9]{2})([0-9]{2,3})[\._ -][^\\/]*$''' % (valid_filename_chars_regex)),
] # contains regex parsing filename parsing strings used to extract info from video filenames

# Episode meta data that is massaged
massage={'writer':'|','director':'|', 'overview':'&', 'gueststars':'|' }
# Keys and titles used for episode data (option '-D')
data_keys	=['seasonnumber','episodenumber','episodename','firstaired','director','overview','rating','writer','filename','language' ]
data_titles	=['Season:','Episode:','Title:','ReleaseDate:','Director:','Plot:','UserRating:','Writers:','Episode Image:','Language:' ]
# High level dictionay keys for select graphics URL(s)
fanart_key='fanart'
banner_key='series'
poster_key='poster'
season_key='season'
# Lower level dictionay keys for select graphics URL(s)
poster_series_key='680x1000'
poster_season_key='season'
fanart_hires_key='1920x1080'
fanart_lowres_key='1280x720'
banner_series_key='graphical'
banner_season_key='seasonwide'
# Type of graphics being requested
poster_type='Poster'
fanart_type='Fanart'
banner_type='Banner'

# Cache directory name specific to the user. This avoids permission denied error with a common cache dirs 
cache_dir="/tmp/tvdb_api_%s/" % os.geteuid() 

def debuglog(message):
	message+='\n'
	target_socket = open(logfile, "a") 
	target_socket.write(message)
	target_socket.close()
	return
# end debuglog

# new Tvdb method to search for a series and return it's sid
def _series_by_sid(self, sid):
	"Lookup a series via it's sid"
	seriesid = 'sid:' + str(sid)
	if not self.corrections.has_key(seriesid):
		self._getShowData(sid)
		self.corrections[seriesid] = sid
	return self.shows[sid]
tvdb_api.Tvdb.series_by_sid = _series_by_sid

# Search for a series by SID or Series name
def search_for_series(tvdb, sid_or_name):
	"Get series data by sid or series name of the Tv show"
	if SID == True:
		return tvdb.series_by_sid(sid_or_name)
	else:
		return tvdb[sid_or_name]
# end search_for_series

# Verify that a Series or Series and Season exists on thetvdb.com
def searchseries(t, opts, series_season_ep):
	global SID
	series_name=''
	if opts.configure != "" and override.has_key(series_season_ep[0].lower()):
		series_name=override[series_season_ep[0].lower()][0] 	# Override series name
	else:
		series_name=series_season_ep[0] 				# Leave the series name alone
	try:
		# Search for the series or series & season or series & season & episode
		if len(series_season_ep)>1:
			if len(series_season_ep)>2: 							# series & season & episode
				seriesfound=search_for_series(t, series_name)[ int(series_season_ep[1]) ][ int(series_season_ep[2]) ]
			else:
				seriesfound=search_for_series(t, series_name)[ int(series_season_ep[1]) ] # series & season
		else:
			seriesfound=search_for_series(t, series_name)	# Series only
	except tvdb_shownotfound:
		# No such show found.
		# Use the show-name from the files name, and None as the ep name
		sys.stderr.write("! Warning: Series (%s) not found\n" % (
			series_name )
		)
		sys.exit(False)
	except (tvdb_seasonnotfound, tvdb_episodenotfound, tvdb_attributenotfound):
		# The season, episode or name wasn't found, but the show was.
		# Use the corrected show-name, but no episode name.
		if len(series_season_ep)>2: 
			sys.stderr.write("! Warning: For Series (%s), season (%s) or Episode (%s) not found \n" % (
				series_name, series_season_ep[1], series_season_ep[2] )
			)
		else:
			sys.stderr.write("! Warning: For Series (%s), season (%s) not found \n" % (
				series_name, series_season_ep[1] )
			)
		sys.exit(False)
	except tvdb_error, errormsg:
		# Error communicating with thetvdb.com
		if SID == True: # Maybe the 5 digit number was a series name (e.g. 90210)
			SID = False
			return searchseries(t, opts, series_season_ep)
		sys.stderr.write(
			"! Warning: Error contacting www.thetvdb.com:\n%s\n" % (errormsg)
		)
		sys.exit(False)
	except tvdb_userabort, errormsg:
		# User aborted selection (q or ^c)
		print "\n", errormsg
		sys.exit(False)
	else:
		if opts.raw==True:
			print "="*20
			print "Raw Series Data:\n"
			if len(series_season_ep)>1:
				print t[ series_name ][ int(series_season_ep[1]) ]
			else:
				print t[ series_name ]
			print "="*20
		return(seriesfound)
# end searchseries

# Retrieve Poster or Fan Art or Banner graphics URL(s)
def get_graphics(t, opts, series_season_ep, graphics_type, single_option, language=''):
	banners='_banners'
	series_name=''
	graphics=[]
	if opts.configure != "" and override.has_key(series_season_ep[0].lower()):
		series_name=override[series_season_ep[0].lower()][0] 	# Override series name
	else:
		series_name=series_season_ep[0] 				# Leave the series name alone

	if graphics_type == fanart_type:		# Series fanart graphics
		lowres_items=[]
		hires_items=[]
		lowres_flag=True
		highres_flag=True
		try:
			lowres_items=search_for_series(t, series_name).data[banners][fanart_key][fanart_lowres_key].values()
		except:
			lowres_flag=False
		try:
			hires_items=search_for_series(t, series_name).data[banners][fanart_key][fanart_hires_key].values()
		except:
			highres_flag=False
		if highres_flag==False and lowres_flag==False:
			return []
		for item in lowres_items:
			graphics.append(item)
		for item in hires_items:
			graphics.append(item)
	elif len(series_season_ep) == 1:
		if graphics_type == banner_type: 		# Series Banners
			try:
				graphics=[b for b in
						search_for_series(t, series_name).data[banners][banner_key][banner_series_key].values()] 
			except:
				return []
		else:									# Series Posters
			try:
				graphics=[b for b in
						search_for_series(t, series_name).data[banners][poster_key][poster_series_key].values()] 
			except:
				return []
	else:
		if graphics_type == banner_type: 		# Season Banners
			try:
				graphics=[b for b in
						search_for_series(t, series_name).data[banners][season_key][banner_season_key].values() 
						if b[season_key] == series_season_ep[1]]
			except:
				return []
		else:									# Season Posters
			try:
				graphics=[b for b in
						search_for_series(t, series_name).data[banners][season_key][poster_season_key].values() 
						if b[season_key] == series_season_ep[1]]
			except:
				return []

	graphicsURLs=[]
	if single_option==False:
		graphicsURLs.append(graphics_type+':')

	for URL in graphics:
		if language != '':		# Is there a language to filter URLs on? 
			if language == URL['language']:
				graphicsURLs.append((URL['_bannerpath']).replace(http_find, http_replace))
		else:
			graphicsURLs.append((URL['_bannerpath']).replace(http_find, http_replace))
	if opts.debug == True:
		print "\nGraphics:\n", graphicsURLs

	if len(graphicsURLs) == 1 and graphicsURLs[0] == graphics_type+':':
		return []						# Due to the language filter there may not be any URLs

	return(graphicsURLs)
# end get_graphics

# Massage episode name to match those in thetvdb.com for this series
def massageEpisode_name(ep_name, series_season_ep):
	for edit in override[series_season_ep[0].lower()][1]:
		ep_name=ep_name.replace(edit[0],edit[1])	# Edit episode name for each set of strings 
	return ep_name
# end massageEpisode_name

# Remove '|' and replace with commas
def change_to_commas(meta_data):
	if not meta_data: return meta_data
	meta_data = (u'|'.join([d for d in meta_data.split('| ') if d]))
	return (u', '.join([d for d in meta_data.split('|') if d]))
# end change_to_commas

# Change &amp; values to ascii equivalents
def change_amp(text):
	if not text: return text
	text = text.replace("&quot;", "'").replace("\r\n", " ")
	text = text.replace(r"\'", "'")
	return text
# end change_amp

# Prepare for includion into a DB
def make_db_ready(text):
	if not text: return text
	text = text.replace(u'\u2013', "-")
	text = text.replace(u'\u2014', "-")
	text = text.replace(u'\u2018', "'")
	text = text.replace(u'\u2019', "'")
	text = text.replace(u'\u2026', "...")
	text = text.replace(u'\u201c', '"')
	text = text.replace(u'\u201d', '"')
	text = text.encode('latin-1', 'backslashreplace')
	return text
# end make_db_ready

# Get Series Episode data by season
def Getseries_episode_data(t, opts, series_season_ep):
	args = len(series_season_ep)
	series_name=''
	if opts.configure != "" and override.has_key(series_season_ep[0].lower()):
		series_name=override[series_season_ep[0].lower()][0] 	# Override series name
	else:
		series_name=series_season_ep[0] 				# Leave the series name alone
	seasons=search_for_series(t, series_name).keys()	# Get the seasons for this series
	for season in seasons:
		if args > 1:						# If a season was specified skip other seasons
			if season != int(series_season_ep[1]):
				continue
		episodes=search_for_series(t, series_name)[season].keys()	# Get the episodes for this season
		for episode in episodes:			# If an episode was specified skip other episodes
			if args > 2:
				if episode != int(series_season_ep[2]):
					continue
			extra_ep_data=[]
			available_keys=search_for_series(t, series_name)[season][episode].keys()
			key_values=[]
			for values in data_keys:	# Initialize an array for each possible data element for 
				key_values.append('')	# each episode within a season
			for key in available_keys:
				try:
					i = data_keys.index(key)	# Include only specific episode data 
				except ValueError:
					if search_for_series(t, series_name)[season][episode][key] != None:
						text = search_for_series(t, series_name)[season][episode][key]
						text = change_amp(text)
						text = change_to_commas(text)
						extra_ep_data.append("%s:%s" % (key.title(), text))
					continue
				text = search_for_series(t, series_name)[season][episode][key]
				if text == None:
					value = str(None)
				else:
					text = change_amp(text)
					value = change_to_commas(text)
				key_values[i]=value
			index = 0
			if SID == False:
				print "Series:%s" % series_name.encode('utf-8')		# Ouput the full series name
			else:
				print "Series:%s" % search_for_series(t, series_name)[u'seriesname'].encode('utf-8')
			for key in data_titles:
				try:
					if key_values[index] != None:
						if data_titles[index] == u'ReleaseDate:' and len(key_values[index]) > 4:
							print '%s%s'% (u'Year:', key_values[index][:4])
						print '%s%s' % (data_titles[index], key_values[index])
				except:
					sys.stdout.write(('%s%s\n' % (data_titles[index], key_values[index])).encode('utf-8'))
				index+=1
			for extra_data in extra_ep_data:
				try:
					print extra_data.encode('utf-8')
				except:
					print "Invalid unicode data"
# end Getseries_episode_data

# Get Series Season and Episode numbers
def Getseries_episode_numbers(t, opts, series_season_ep):
	series_name=''
	ep_name=''
	if opts.configure != "" and override.has_key(series_season_ep[0].lower()):
		series_name=override[series_season_ep[0].lower()][0] 		# Override series name
		ep_name=series_season_ep[1]
		if len(override[series_season_ep[0].lower()][1]) != 0:		# Are there search-replace strings?
			ep_name=massageEpisode_name(ep_name, series_season_ep)
	else:
		series_name=series_season_ep[0] 				# Leave the series name alone
		ep_name=series_season_ep[1]						# Leave the episode name alone

	season_ep_num=search_for_series(t, series_name).search(ep_name)
	if len(season_ep_num) != 0:
		for episode in season_ep_num:
			if episode['episodename'].lower() == ep_name.lower():
				print season_and_episode_num % (int(episode['seasonnumber']), int(episode['episodenumber']))
				break
# end Getseries_episode_numbers

# Set up a custom interface to get all series matching a partial series name
class returnAllSeriesUI(tvdb_ui.BaseUI):
	def __init__(self, config, log):
		self.config = config
		self.log = log

	def selectSeries(self, allSeries):
		return allSeries
# ends returnAllSeriesUI

def initialize_override_dictionary(useroptions):
	"""	Change variables through a user supplied configuration file
	return False and exit the script if there are issues with the configuration file values
	"""
	if useroptions[0]=='~':
		useroptions=os.path.expanduser("~")+useroptions[1:]
	if os.path.isfile(useroptions) == False:
		sys.stderr.write(
			"! The specified user configuration file (%s) is not a file\n" % useroptions
		)
		sys.exit(False)
	massage = {}
	overrides = {}
	cfg = ConfigParser.SafeConfigParser()
	cfg.read(useroptions)

	for section in cfg.sections(): 
		if section == 'regex':
			# Change variables per user config file
			for option in cfg.options(section): 
				name_parse.append(re.compile(cfg.get(section, option) % (valid_filename_chars_regex)))
			continue
		if section =='ep_name_massage':
			for option in cfg.options(section):
				tmp =cfg.get(section, option).split(',')
				if len(tmp)%2 and len(cfg.get(section, option)) != 0:
					sys.stderr.write("! For (%s) 'ep_name_massage' values must be in pairs\n" % option)
					sys.exit(False)
				tmp_array=[]
				i=0
				while i != len(tmp):
					tmp_array.append([tmp[i].replace('"',''), tmp[i+1].replace('"','')])
					i+=2
				massage[option]=tmp_array
			continue
		if section =='series_name_override':
			for option in cfg.options(section): 
				overrides[option] = cfg.get(section, option)
			tvdb = tvdb_api.Tvdb(banners=False, debug = False, interactive = False, cache = cache_dir, custom_ui=returnAllSeriesUI)
			for key in overrides.keys():
				sid = overrides[key]
				if len(sid) == 0:
					continue
				try:					# Check that the SID (Series id) is numeric
					dummy = int(sid)
				except:
					sys.stdout.write("! Series (%s) Invalid SID (not numeric) [%s]\n" % (key, sid))
					sys.exit(False)
				# Make sure that the series name is not empty or all blanks
				if len(key.replace(' ','')) == 0:
					sys.stdout.write("! Invalid Series name (must have some non-blank characters) [%s]\n" % key)
					print parts
					sys.exit(False)

				try:
					series_name_sid=tvdb.series_by_sid(sid)
				except:
					sys.stdout.write("! Invalid Series (no matches found in thetvdb,com) (%s) sid (%s)\n" % (key, sid))
					sys.exit(False)
				overrides[key]=series_name_sid[u'seriesname'].encode('utf-8')
			continue
	
	for key in overrides.keys():
		override[key] = [overrides[key],[]]

	for key in massage.keys():
		if override.has_key(key):
			override[key][1]=massage[key]
		else:
			override[key]=[key, massage[key]]
	return
# END initialize_override_dictionary

def main():
	parser = OptionParser(usage="%prog usage: ttvdb -hdruvioMPFBD [parameters]\n <series name or 'series and season number' or 'series and season number and episode number'>")

	parser.add_option(  "-d", "--debug", action="store_true", default=False, dest="debug",
                        help="Show debugging info")
	parser.add_option(  "-r", "--raw", action="store_true",default=False, dest="raw", 
                        help="Dump raw data only")
	parser.add_option(  "-u", "--usage", action="store_true", default=False, dest="usage",
                        help="Display examples for executing the ttvdb script")
	parser.add_option(  "-v", "--version", action="store_true", default=False, dest="version",
                        help="Display version and author")
	parser.add_option(  "-i", "--interactive", action="store_true", default=False, dest="interactive",
                        help="Interaction mode (allows selection of a specific Series)")
	parser.add_option(  "-c", "--configure", metavar="FILE", default="", dest="configure",
                        help="Use configuration settings")
	parser.add_option(  "-l", "--language", action="store_true", default=False, dest="language",
                        help="Only select graphics that match the local language (e.g. 'en')")
	parser.add_option(  "-n", "--num_seasons", action="store_true", default=False, dest="num_seasons",
                        help="Return the season numbers for a series")
	parser.add_option(  "-t", "--toprated", action="store_true", default=False, dest="toprated",
                        help="Only return the top rated graphics for a TV Series")
	parser.add_option(  "-m", "--mythvideo", action="store_true", default=False, dest="mythvideo",
                        help="Conform to mythvideo standards when processing -M, -P, -F and -D")
	parser.add_option(  "-M", "--list", action="store_true", default=False, dest="list",
                        help="(Not ACTIVE) Get Series list")
	parser.add_option(  "-P", "--poster", action="store_true", default=False, dest="poster",
                        help="Get Series Poster URL")
	parser.add_option(  "-F", "--fanart", action="store_true", default=False, dest="fanart",
                        help="Get Series fan art URL(s)")
	parser.add_option(  "-B", "--backdrop", action="store_true", default=False, dest="banner",
                        help="Get Series banner/backdrop URL(s)")
	parser.add_option(  "-D", "--data", action="store_true", default=False, dest="data",
                        help="Get Series data")
	parser.add_option(  "-N", "--numbers", action="store_true", default=False, dest="numbers",
                        help="Get Season and Episode numbers")

	opts, series_season_ep = parser.parse_args()

	if opts.debug == True:
		print "opts", opts
		print "\nargs", series_season_ep

	# Process version command line requests
	if opts.version == True:
		sys.stdout.write("\nTitle: (%s); Version: (%s); Author: (%s)\n" % (
		__title__, __version__, __author__ ))
		sys.stdout.write("This python script uses the included python module tvdb_api.py (v0.6 or higher). Also found at\n\
'http://pypi.python.org/pypi?%3Aaction=search&term=tvnamer&submit=search' thanks to the authors 'dbr/Ben' for this excellant module.\n\n\
The tvdb_api.py module uses the full access XML api published by thetvdb.com see:\n\
'http://thetvdb.com/wiki/index.php?title=Programmers_API'\n\n\
Users of this script are encouraged to populate thetvdb.com with TV show information, posters,\n\
fan art and banners. The richer the source the more valuable the script.\n\n"
		)
		sys.exit(True)

	# Process usage command line requests
	if opts.usage == True:
		sys.stdout.write(usage_txt)
		sys.exit(True)

	if len(series_season_ep) == 0:
		parser.error("! No series or series season episode supplied")
		sys.exit(False)

	# Default output format of season and episode numbers
	global season_and_episode_num
	season_and_episode_num='S%02dE%02d'			# Format output example "S04E12"

	if opts.numbers == False:
		if len(series_season_ep) > 1:
			if series_season_ep[1].lstrip('0123456789') != "":
				parser.error("! Season is not numeric")
				sys.exit(False)
		if len(series_season_ep) > 2:
			if series_season_ep[2].lstrip('0123456789') != "":
				parser.error("! Episode is not numeric")
				sys.exit(False)
	else:
		if len(series_season_ep) < 2:
			parser.error("! An Episode name must be included")
			sys.exit(False)
		if len(series_season_ep) == 3:
			season_and_episode_num = series_season_ep[2]	# Override default output format

	if opts.debug == True:
		print series_season_ep

	if opts.debug == True:
		print "#"*20
		print "# series_season_ep array(",series_season_ep,")"
	
	if opts.debug == True:
		print "#"*20
		print "# Starting tvtvb"
		print "# Processing (%s) Series" % ( series_season_ep[0] )

	# Access thetvdb.com API with banners (Posters, Fanart, banners) data retrieval enabled
	if opts.list ==True: 
		t = tvdb_api.Tvdb(banners=False, debug = opts.debug, cache = cache_dir, custom_ui=returnAllSeriesUI)
	elif opts.interactive == True:
		t = tvdb_api.Tvdb(banners=True, debug=opts.debug, interactive=True,  select_first=False, cache=cache_dir)
	else:
		t = tvdb_api.Tvdb(banners=True, debug = opts.debug, cache = cache_dir)

	# Determine if there is a SID or a series name to search with
	global SID
	SID = False
	if len(series_season_ep[0]) == 5:
		try:					# Check that the SID (Series id)/series name is numeric
			dummy = int(series_season_ep[0])
			SID = True
		except:
			SID = False

	if opts.debug == True:
		print "# ..got tvdb mirrors"
		print "# Start to process series or series_season_ep"
		print "#"*20

	global override
	override={}			# Initialize series name override dictionary
	# If the use wants Series name overrides and a override file exists then create an overide dictionary
	if opts.configure != '':
		if opts.configure[0]=='~':
			opts.configure=os.path.expanduser("~")+opts.configure[1:]
		if os.path.exists(opts.configure) == 1: # Do overrides exist?
			initialize_override_dictionary(opts.configure)
		else:
			debuglog("! The specified override file (%s) does not exist" % opts.configure)
			sys.exit(False)
	if len(override) == 0:
		opts.configure = False	# Turn off the override option as there is nothing to override

	# Check if a video name was passed and if so parse it for series name, season and episode numbers
	if len(series_season_ep) == 1: 
		for r in name_parse:
			match = r.match(series_season_ep[0])
			if match:
				seriesname, seasno, epno = match.groups()
				#remove ._- characters from name (- removed only if next to end of line)
				seriesname = re.sub("[\._]|\-(?=$)", " ", seriesname).strip()
				series_season_ep = [seriesname, str(int(seasno)), str(int(epno))]
				break # Matched - to the next file!

	# Fetch a list of matching series names 
	if (opts.list ==True ): 
		allSeries=t._getSeries(series_season_ep[0])
		for series_name_sid in allSeries: # list search results
			print "%s:%s" % (series_name_sid['sid'], series_name_sid['name'])
		sys.exit(True) # The Series list option (-M) is the only option honoured when used

	# Verify that thetvdb.com has the desired series_season_ep. 
	# Exit this module if series_season_ep is not found
	if opts.numbers == False and opts.num_seasons == False:
		seriesfound=searchseries(t, opts, series_season_ep)
		x=1
	else:
		x=[]
		x.append(series_season_ep[0])	# Only use series name in check
		seriesfound=searchseries(t, opts, x)

	# Return the season numbers for a series
	if opts.num_seasons == True:
		season_numbers=''
		for x in seriesfound.keys():	
			season_numbers+='%d,' % x
		print season_numbers[:-1]
		sys.exit(True) # Option (-n) is the only option honoured when used

	# Dump information accessable for a Series and ONLY first season of episoded data
	if opts.debug == True:
		print "#"*20
		print "# Starting Raw keys call"
		print "Lvl #1:" 								# Seasons for series
		x = t[series_season_ep[0]].keys()
		print t[series_season_ep[0]].keys()
		print "#"*20
		print "Lvl #2:"									# Episodes for each season
		for y in x:
			print t[series_season_ep[0]][y].keys()
		print "#"*20
		print "Lvl #3:"									# Keys for each episode within the 1st season
		z = t[series_season_ep[0]][1].keys()
		for aa in z:
			print t[series_season_ep[0]][1][aa].keys()
		print "#"*20
		print "Lvl #4:"									# Available data for each episode in 1st season
		for aa in z:
			codes = t[series_season_ep[0]][1][aa].keys()
			print "\n\nStart:"
			for c in codes:
				print "="*50
				print 'Key Name=('+c+'):'
				print t[series_season_ep[0]][1][aa][c]
				print "="*50
		print "#"*20
		sys.exit (True)

	if opts.numbers == True:				# Fetch and output season and episode numbers
		Getseries_episode_numbers(t, opts, series_season_ep)
		sys.exit(True) # The Numbers option (-N) is the only option honoured when used

	if opts.data == True:				# Fetch and output episode data
		if opts.mythvideo:
			if len(series_season_ep) != 3:
				print "Season and Episode numbers required."
			else:
				Getseries_episode_data(t, opts, series_season_ep)
		else:
			Getseries_episode_data(t, opts, series_season_ep)

	# Find the local language
	local_language=''
	if opts.language == True:
		lang = locale.getdefaultlocale()
		local_language=lang[0][:2]

	# Fetch the requested graphics URL(s)
	if opts.debug == True:
		print "#"*20
		print "# Checking if Posters, Fanart or Banners are available"
		print "#"*20

	if opts.configure != "" and override.has_key(series_season_ep[0].lower()):
		banners_keys = search_for_series(t, override[series_season_ep[0].lower()][0])['_banners'].keys()
	else:
		banners_keys = search_for_series(t, series_season_ep[0])['_banners'].keys()
	banner= False
	poster= False
	fanart= False

	for x in banners_keys:				# Determine what type of graphics is available
		if x == fanart_key:
			fanart=True
		elif x== poster_key:
			poster=True
		elif x==season_key or x==banner_key:
			banner=True

	# Make sure that some graphics URL(s) (Posters, FanArt or Banners) are available
	if ( fanart!=True and poster!=True and banner!=True ):
		sys.exit(True)
	
	if opts.debug == True:
		print "#"*20
		print "# One or more of Posters, Fanart or Banners are available"
		print "#"*20

	# Determine if graphic URL identification output is required	
	y=0
	single_option=True
	if opts.poster==True:
		y+=1
	if opts.fanart==True:
		y+=1
	if opts.banner==True:
		y+=1
	if y > 1:							
		single_option=False

	# Determine if only top rated by thetvdb.com graphics has been requested:
	if opts.toprated == True:
		series_name=''
		if opts.configure != "" and override.has_key(series_season_ep[0].lower()):
			series_name=override[series_season_ep[0].lower()][0] 	# Override series name
		else:
			series_name=series_season_ep[0] 				# Leave the series name alone
		if opts.poster==True:
			if search_for_series(t, series_name)['poster'] != None:
				if single_option==True:
					print (search_for_series(t, series_name)['poster']).replace(http_find, http_replace)
				else:
					print "Poster:%s" % (search_for_series(t, series_name)['poster']).replace(http_find, http_replace)
		if opts.fanart==True:
			if search_for_series(t, series_name)['fanart'] != None:
				if single_option==True:
					print (search_for_series(t, series_name)['fanart']).replace(http_find, http_replace)
				else:
					print "Fanart:%s" % (search_for_series(t, series_name)['fanart']).replace(http_find, http_replace)
		if opts.banner==True:
			if search_for_series(t, series_name)['banner'] != None:
				if single_option==True:
					print (search_for_series(t, series_name)['banner']).replace(http_find, http_replace)
				else:
					print "Banner:%s" % (search_for_series(t, series_name)['banner']).replace(http_find, http_replace)
		sys.exit(True)					# Only the top rated for a TV Series is returned

	if (poster==True and opts.poster==True and opts.raw!=True):	# Get posters and send to stdout
		season_poster_found = False
		if opts.mythvideo:
			if len(series_season_ep) < 2:
				print "Season and Episode numbers required."
				sys.exit(True)			
		for p in get_graphics(t, opts, series_season_ep, poster_type, single_option, local_language):
			print p
			season_poster_found = True
		if season_poster_found == False: # If there were no season posters get the series top poster
			series_name=''
			if opts.configure != "" and override.has_key(series_season_ep[0].lower()):
				series_name=override[series_season_ep[0].lower()][0] 	# Override series name
			else:
				series_name=series_season_ep[0] 				# Leave the series name alone
			print (search_for_series(t, series_name)['poster']).replace(http_find, http_replace)

	if (fanart==True and opts.fanart==True and opts.raw!=True):	# Get Fan Art and send to stdout
		for f in get_graphics(t, opts, series_season_ep, fanart_type, single_option, local_language):
			print f

	if (banner==True and opts.banner==True and opts.raw!=True):	# Get Banners and send to stdout
		if opts.mythvideo:
			if len(series_season_ep) < 2:
				print "Season and Episode numbers required."
				sys.exit(True)			
		for b in get_graphics(t, opts, series_season_ep, banner_type, single_option, local_language):
			print b

	if opts.debug == True:
		print "#"*20
		print "# Processing complete"
		print "#"*20
	sys.exit(True)
#end main

if __name__ == "__main__":
	main()
